module Drawing where

import Graphics.HGL
import Array (Array, (!))
import Prelude hiding (Left, Right)
import Maybe (mapMaybe, fromMaybe)

import Board
import State


-- Size parameters

-- Size of the window
winX :: Int
winX = 620
winY :: Int
winY = 550

-- Border around maze
mbX :: Int
mbX = 40
mbY :: Int
mbY = 40

-- Size of rows and colums, i.e. height of row and width of columns
rowSize :: Int
rowSize = (winY- 2*mbY- 50) `div` (mazeRows- 1)
colSize :: Int
colSize = (winX- 2*mbX) `div` (mazeCols- 1)


-- Size of the player
pS :: Int
pS = 30


-- Utility functions

dirToAngle :: Dir-> Angle
dirToAngle Right = 0
dirToAngle Up    = 90
dirToAngle Left  = 180
dirToAngle Down  = 270
  

circle :: Point-> Int-> Graphic
circle (px, py) r = ellipse (px- r, py- r) (px+ r, py+ r)

rect :: Point-> Point-> Graphic
rect (px, py) (qx, qy) = polygon [(px, py), (px, qy), (qx, qy), (qx, py)]

-- Maze utility functions
mazeOffset :: Point-> Point
mazeOffset (px, py)= (px+ mbX, py+ mbY)

dotPos :: Dot-> (Int, Int)
dotPos d = 
  let (r, c)= pos d 
  in  mazeOffset (colSize* c, rowSize* r)

poseToPoint :: Pose-> Point
poseToPoint (Pose{from= f, to= t, dist= d}) =
  let (fx, fy) = dotPos f
      (tx, ty) = dotPos t
  in  if f == t then (fx, fy)
      else (round(fromIntegral tx+ d* (fromIntegral(fx- tx))),
	    round(fromIntegral ty+ d* (fromIntegral(fy- ty))))
  

-- Draw the player
drawPlayer :: Pose-> Graphic
drawPlayer pp = 
  let (px, py) = poseToPoint pp
  in withColor Yellow $
      arc (px- pS `div` 2, py- pS `div` 2)  
          (px+ pS `div` 2, py+ pS `div` 2)
          (dirToAngle (fromMaybe Up (heading pp))+ 45) 270


-- Draw the background (i.e. the maze)
drawBackground :: Graphic
drawBackground = 
  let dotsto d = zip (repeat d) (mapMaybe (maze d) allDirs)
      alldots  = concatMap dotsto allDots
      lines    = map (\ (d1, d2)-> (dotPos d1, dotPos d2)) alldots
  in  overGraphics $ map (withColor Blue. uncurry line) lines

-- Draw the maze state (i.e. the pills)
drawMaze :: Array Dot DotState-> Graphic
drawMaze field = 
 let drawState d = 
       case field ! d of 
         MtDot -> emptyGraphic
         WeePill -> withRGB (RGB 255 251 153) (circle (dotPos d) 5)
         BigPill -> withRGB (RGB 255 251 57) (circle (dotPos d) 10)
 in  overGraphics $ map drawState allDots
  

-- Draw statistics
drawStats :: Int-> Int-> Int-> Graphic
drawStats score lives pill = 
  let sy   = winY- (mbY+ 20)
      size = (((winX `div` 2)- mbX) * pill) `div` pillTime
  in  overGraphics 
      [ withColor Green (text (mbX, sy) $ "Score: "++ show score)
      , withColor Blue  (text (mbX+ winX `div` 4, sy) $ "Nerves: "++ concat (replicate lives "* "))
      , withColor Blue (rect (mbX+ winX `div` 2, sy- 10)
			     (mbX+ winX `div` 2+ size, sy+ 10))
      ]

-- Draw the whole state
draw :: State-> Graphic
draw s = overGraphics $ reverse
  [ drawBackground
  , drawMaze (mazest s)
  , drawStats (score s) (lives s) (pill s)
  , drawPlayer (playerpose s){heading= playerheading s}
  ]
