module State where

import Array(Array, (//), (!))
import Random(randomRIO)
import Maybe (mapMaybe, fromMaybe, listToMaybe, maybeToList)

import Debug.Trace(trace)

import Board


data Ghost = Blinky | Pinky | Inky | Clyde deriving (Eq, Ord, Enum, Show)

ghosts :: [Ghost]
ghosts = [Blinky .. Clyde]

ghostScore :: Ghost-> Int
ghostScore Blinky = 200
ghostScore Pinky  = 250
ghostScore Inky   = 300
ghostScore Clyde  = 500

initialPlayerPose :: Pose
initialPlayerPose = Pose { from = playerStart
			 , to = playerStart
			 , dist = 0
			 , heading = Nothing
			 }

-- A pose: whence we come, where we go, how far we've gone, and where we're headed.
data Pose = Pose { from :: Dot
		 , to   :: Dot
		 , dist :: Double
		 , heading :: Maybe Dir
		 }  deriving Show

-- An eaten power pill lasts that many cycles
pillTime :: Int
pillTime = 30* 5 -- i.e. 5 seconds

-- Player speed
playerSpeed :: Double
playerSpeed = 0.05


-- Distances smaller than this are considered "0"
minDist :: Double 
minDist = 0.001

-- Distances smaller than this are supposed to touch. Depends a bit 
-- the graphics (i.e. how large are ghosts/player in rel to distance
-- between dots)
touchDist :: Double
touchDist = 0.01

-- Move a pose
movePose :: Pose-> Double->Pose
movePose p v
  | from p == to p = p -- not moving
  | otherwise = let d' = dist p - v
                    p' = p {dist= d'}
                in if d' <= minDist 
		   then p'{from= to p, dist= 0} -- reached destination
                   else p'

-- Check wether two poses touch.
-- This is the case if either they have the same destination and origin, 
-- and the same distance (up to touchDist above), or if the have the 
-- opposite destination and origin, and an inverse distance.
posesTouch :: Pose-> Pose-> Bool
posesTouch p q =
     (from p == from q && to p == to q && abs (dist p- dist q) < touchDist)
  || (from p == to q && to p == from q && abs ((1- dist p)- dist q) < touchDist)

-- The global state of the game
data State = State { playerpose :: Pose
		   , playerheading :: Maybe Dir -- direction to take next
		   , pill   :: Int
		   , lives  :: Int
		   , dead   :: Bool
		   , score  :: Int
		   , mazest  :: Array Dot DotState
	           } 

pickR :: [a]-> IO a
pickR [] = error "State.pickR: empty list"
pickR s  = do n <- randomRIO (1, length s)
              return $ s !! (n-1)


-- ----------------------------------------------------------------------
--
-- Move around the player
--
-- ----------------------------------------------------------------------

-- Moving the player, we have three cases:
-- we have reached the end of the current transition (case (i))
-- so we can take any allowed direction.
-- if we are in a current transition, we can only reverse (case (ii)
-- otherwise we just continue (case (iii))
movePlayer :: Maybe Dir-> Pose-> Pose
movePlayer md p 
  | from p == to p =  -- case (i)
      let h' = case md of Nothing -> heading p
                          Just md-> Just md
          wh = case h' of Nothing -> Nothing
                          Just d-> maze (from p) d
      in  case wh of
            Just t -> let nupose= Pose { from = from p
                                       , to = t
                                       , dist = 1
                                       , heading= h'
                                       }
                      in  movePose nupose playerSpeed
            Nothing-> Pose { from= to p
                           , to = to p
                           , dist= 0
                           , heading= Nothing
                           }
  | opposite md (heading p) =  -- case (ii)
      let nupose = Pose { from= to p
                        , to= from p
                        , dist= 1- dist p
                        , heading= md
                        }
      in movePose nupose playerSpeed
  | otherwise = movePose p playerSpeed -- case (iii)

-- ----------------------------------------------------------------------
--
-- Calculate next state
--
-- ----------------------------------------------------------------------

nextState :: State-> Maybe Dir-> IO State
nextState s nh =
  do let ph  = -- nh overrides playerheading s, if given
               listToMaybe (maybeToList nh ++ maybeToList (playerheading s))
         pp' = movePlayer ph (playerpose s)
     let eaten  = if (from pp' == to pp') then Just (mazest s!from pp') else Nothing
         maze'  = case eaten of Nothing -> mazest s
                                Just _  -> (mazest s) // [(from pp', MtDot)]
         pill'  = case eaten of Just BigPill -> pillTime
                                _            -> max 0 (pill s- 1)
         sc'    = score s + case eaten of Just BigPill -> 50
                                          Just WeePill -> 10
                                          _ -> 0
     return (State { playerpose = pp'
		   , playerheading = ph
		   , pill= pill'
		   , dead= dead s
		   , lives= lives s
		   , score= sc'
		   , mazest= maze' 
		   })

-- Start a new game (reduces number of lives by one)
newGame :: State-> State
newGame s= 
  State{ playerpose = initialPlayerPose
       , playerheading = Nothing
       , pill = 0
       , dead = False
       , score = score s
       , lives = lives s- 1
       , mazest = initialmaze
       }

-- Initial State
initialState :: State
initialState = 
  newGame $ State { score = 0
		  , lives = 4 }
